/*
 * Copyright (C) 2006-2007 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */
/*
 * based in parts on libfaad, Copyright (C) 2003-2005 M. Bakker,
 * Ahead Software AG
 */

/** \file aac_afh.c para_server's aac audio format handler */

#include "server.cmdline.h"
#include "server.h"
#include "vss.h"
#include "afh.h"
#include "error.h"
#include "string.h"
#include "aac.h"
#include "fd.h"

/** size of the input buffer, must be big enough to hold header */
#define DEFAULT_INBUF_SIZE 65536

static FILE *infile;
static int inbuf_size;
static unsigned char *inbuf;
static size_t inbuf_len;
static struct audio_format_handler *af;
static size_t num_chunks;
static size_t entry;

static size_t *chunk_table;
static NeAACDecHandle handle;

static void aac_close_audio_file(void)
{
	if (!infile)
		return;
	fclose(infile);
	infile = NULL;
}

static int aac_find_stsz(unsigned char *buf, unsigned buflen, size_t *skip)
{
	int i;

	for (i = 0; i + 16 < buflen; i++) {
		unsigned char *p = buf + i;
		unsigned sample_count, sample_size;

		if (p[0] != 's' || p[1] != 't' || p[2] != 's' || p[3] != 'z')
			continue;
		PARA_INFO_LOG("found stsz@%d\n", i);
		i += 8;
		sample_size = aac_read_int32(buf + i);
		PARA_INFO_LOG("sample size: %d\n", sample_size);
		i += 4;
		sample_count = aac_read_int32(buf + i);
		i += 4;
		PARA_INFO_LOG("sample count: %d\n", sample_count);
		*skip = i;
		return sample_count;
	}
	PARA_WARNING_LOG("stsz not found, buflen: %d\n", buflen);
	return -E_STSZ;
}

static int read_chunk_table(size_t skip)
{
	int ret, i;
	size_t sum = 0;

	for (;;) {
		ret = aac_find_stsz(inbuf, inbuf_len, &skip);
		if (ret >= 0)
			break;
		ret = read(fileno(infile), inbuf, inbuf_size);
		if (ret <= 0)
			return -E_AAC_READ;
		PARA_INFO_LOG("next buffer: %d bytes\n", ret);
	}
	num_chunks = ret;
	PARA_INFO_LOG("sz table has %zu entries\n", num_chunks);
	free(chunk_table);
	chunk_table = para_malloc(num_chunks * sizeof(size_t));
	for (i = 0; i < num_chunks; i++) {
		if (skip + 4 > inbuf_len) {
			skip = inbuf_len - skip;
			memmove(inbuf, inbuf + inbuf_len - skip, skip);
			ret = read(fileno(infile), inbuf + skip, inbuf_size - skip);
			if (ret <= 0)
				return -E_AAC_READ;
			inbuf_len = ret + skip;
			skip = 0;
			PARA_INFO_LOG("next buffer: %zu bytes\n", inbuf_len);
		}
		sum += aac_read_int32(inbuf + skip);
		chunk_table[i] = sum;
		skip += 4;
		if (i < 10 || i + 10 > num_chunks)
			PARA_DEBUG_LOG("offset #%d: %zu\n", i, chunk_table[i]);
	}
	return 1;
}

static long unsigned aac_set_chunk_tv(mp4AudioSpecificConfig *mp4ASC)
{
	float tmp = mp4ASC->sbr_present_flag == 1? 2047 : 1023,
		ms = 1000.0 * num_chunks * tmp / mp4ASC->samplingFrequency;
	struct timeval total;

	ms2tv(ms, &total);
	tv_divide(num_chunks, &total, &af->chunk_tv);
	PARA_INFO_LOG("%luHz, %fs (%zd x %lums)\n",
		mp4ASC->samplingFrequency, ms / 1000,
		num_chunks, tv2ms(&af->chunk_tv));
	return ms / 1000;
}

/*
 * Init m4a file and write some tech data to given pointers.
 */
static int aac_get_file_info(FILE *file, char *info_str, long unsigned *frames,
	int *seconds)
{
	int ret, decoder_len;
	size_t skip;
	unsigned long rate = 0;
	unsigned char channels = 0;
	mp4AudioSpecificConfig mp4ASC;

	free(inbuf);
	inbuf_size = DEFAULT_INBUF_SIZE;
	inbuf = para_malloc(inbuf_size);
	infile = file;

	ret = read(fileno(infile), inbuf, inbuf_size);
	if (ret <= 0)
		return -E_AAC_READ;
	inbuf_len = ret;
	ret = aac_find_esds(inbuf, inbuf_len, &skip);
	if (ret < 0)
		return ret;
	decoder_len = ret;
	handle = aac_open();
	ret = NeAACDecInit(handle, inbuf + skip,
		decoder_len, &rate, &channels);
	if (ret < 0)
		return -E_AACDEC_INIT;
	skip += ret;
	PARA_INFO_LOG("rate: %lu, channels: %d\n", rate, channels);
	ret = NeAACDecAudioSpecificConfig(inbuf + skip, inbuf_len - skip,
		&mp4ASC);
	if (ret < 0)
		return -E_MP4ASC;
	ret = read_chunk_table(skip);
	if (ret < 0)
		return ret;
	*frames = num_chunks;
	*seconds = aac_set_chunk_tv(&mp4ASC);
	for (;;) {
		ret = aac_find_entry_point(inbuf, inbuf_len, &skip);
		if (ret >= 0)
			break;
		ret = read(fileno(infile), inbuf, inbuf_size);
		if (ret <= 0)
			return -E_AAC_READ;
		PARA_INFO_LOG("next buffer: %d bytes\n", ret);
	}
	entry = ret;
	sprintf(info_str, "audio_file_info1:%zu x %lums\n"
		"audio_file_info2:\n"
		"audio_file_info3:\n",
		num_chunks,
		tv2ms(&af->chunk_tv));
	tv_scale(20, &af->chunk_tv, &af->eof_tv);
	return 1;
}

/*
 * nothing to do as we'll seek to the correct offset in aac read_chunk() anyway
 */
static int aac_reposition_stream(__a_unused long unsigned request)
{
	return 1;
}

static char *aac_read_chunk(long unsigned current_chunk, ssize_t *len)
{
	int ret;
	size_t pos;

	*len = 0;
	if (current_chunk >= num_chunks)
		return NULL;
	if (!current_chunk) {
		*len = chunk_table[0];
		pos = entry;
	} else {
		*len = chunk_table[current_chunk] - chunk_table[current_chunk - 1];
		pos = entry + chunk_table[current_chunk - 1];
	}
	if (inbuf_size < *len) {
		inbuf = para_realloc(inbuf, *len);
		inbuf_size = *len;
	}
//	PARA_DEBUG_LOG("reading chunk #%lu@%zd (%zd bytes)\n", current_chunk,
//		pos, *len);
	ret = fseek(infile, pos, SEEK_SET);
	if (ret < 0)
		return NULL;
	ret = para_fread(inbuf, *len, 1, infile);
	if (ret != *len)
		return NULL;
//	PARA_DEBUG_LOG("ret: %d, inbuf[0]: %lx - %lx\n", ret, (long unsigned) inbuf[0],
//		(long unsigned) inbuf[4]);
	return (char *)inbuf;
}

static const char* aac_suffixes[] = {"m4a", "mp4", NULL};
/** the init function of the aac audio format handler */
void aac_afh_init(struct audio_format_handler *p)
{
	af = p;
	af->reposition_stream = aac_reposition_stream;
	af->get_file_info = aac_get_file_info,
	af->read_chunk = aac_read_chunk;
	af->close_audio_file = aac_close_audio_file;
	af->get_header_info = NULL;
	af->suffixes = aac_suffixes;
}
